/*
 * soft.c -- supporting the software-timed device /dev/adlS7200
 *
 * Copyright (C) 1998   rubini@linux.it (Alessandro Rubini)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#define __NO_VERSION__
#include <linux/module.h>

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/bios32.h> /* pcibios_* */
#include <linux/pci.h>
#include <linux/delay.h>

#include <asm/system.h>   /* cli(), *_flags */
#include <asm/segment.h>  /* memcpy and such */
#include <asm/io.h>

#include "sysdep-2.1.h"
#include "adl7200.h"        /* local definitions */


/*--------------------------------------------------------------- lseek */

/* That's easy */

lseek_t adl_sft_lseek(struct inode *inode, struct file *file,
		      lseek_off_t offset, int whence)
{
    return -ESPIPE;
}



/*--------------------------------------------------------------- open */
int adl_sft_open (struct inode *inode, struct file *filp)
{
    Adl_Dev *dev = filp->private_data; /* set by adl_open() */

    if (!dev->softbuffer) {
	dev->softbuffer = __get_free_pages(GFP_KERNEL,ADL_SFT_ORDER,0);
	dev->rdhead = dev->rdtail =
	    dev->wrhead = dev->wrtail = dev->softbuffer;
    }
    if (!dev->softbuffer)
	return -ENOMEM;

    /* usage count is managed in the main open() function */
    return 0;
}


/*--------------------------------------------------------------- close */

release_t adl_sft_close (struct inode *inode, struct file *filp)
{
    Adl_Dev *dev = filp->private_data;

    if (!ADL_SFT_WEMPTY(dev)) {
	interruptible_sleep_on(&dev->closeq);
	if (current->signal & ~current->blocked) /* a signal arrived */
	    release_return (-ERESTARTSYS); /* tell the fs layer to handle it */
    }
    dev->usage--;
    if (dev->usage == 0) {
	adl_shutdown(dev);
    }
    MOD_DEC_USE_COUNT;
    release_return(0);
}

/*-------------------------------------------------------- read */

read_write_t adl_sft_read (struct inode *inode, struct file *filp,
                char *buf, count_t count)
{
    Adl_Dev *dev = filp->private_data;

    if (dev->flags & ADL_FLAG_WRITING)
	return -EBUSY;

    return -ENODEV; /* FIXME: read soft */
}

/*-------------------------------------------------------- write */

read_write_t adl_sft_write (struct inode *inode, struct file *filp,
                const char *buf, count_t count)
{
    Adl_Dev *dev = filp->private_data;

    if (dev->flags & ADL_FLAG_READING)
	return -EBUSY;

    return -ENODEV; /* FIXME: write soft */
}

/*-------------------------------------------------------- select (poll) */

#ifdef __USE_OLD_SELECT__
int adl_sft_poll (struct inode *inode, struct file *filp,
                  int mode, select_table *table)
{
    Adl_Dev *dev = filp->private_data;

    if (mode==SEL_IN) {
        if ( (dev->flags & ADL_FLAG_READING) && !ADL_SFT_REMPTY(dev))
	    return 1; /* readable */
        select_wait (&dev->inq, table);
        return 0;
    }
    if (mode==SEL_OUT) {
        if ( (dev->flags & ADL_FLAG_WRITING)  && !ADL_SFT_WFULL(dev))
	return 1; /* writable */
        select_wait (&dev->outq, table);
        return 0;
    }
    return 0; /* not-exceptionable */
}
#else
unsigned int adl_sft_poll (struct file *filp, poll_table *table)
{
    Adl_Dev *dev = filp->private_data;
    int ret=0;

    poll_wait(&dev->inq,  table);
    poll_wait(&dev->outq, table);

    if ((dev->flags & ADL_FLAG_READING) && !ADL_SFT_REMPTY(dev))
        ret |= POLLIN | POLLRDNORM;
    if ((dev->flags & ADL_FLAG_WRITING) && !ADL_SFT_WFULL(dev))
        ret |= POLLOUT | POLLWRNORM;
    return ret;
}
#endif


/*-------------------------------------------------------- fops */

struct file_operations adl_sft_fops = {
    adl_sft_lseek,
    adl_sft_read,
    adl_sft_write,
    NULL,            /* adl_readdir */
    adl_sft_poll,
    NULL,            /* adl_ioctl */
    NULL,            /* adl_mmap */
    adl_sft_open,
    adl_sft_close,
    NULL,          /* nothing more... */
};

